#!/usr/bin/ruby
# SGT  Simple 2D Drawing Library with OpenGL and Ruby/SDL
#
# Copyright (C) 2005 Obayashi Ippei
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is fur-
# nished to do so, subject to the following conditions:
#  
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FIT-
# NESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE X
# CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
require 'enumerator'
require 'opengl'
require 'sdl'
require 'forwardable'

module SDL
  class Surface
    def copy_rect(x,y,w,h)
      flagbase=SDL::SWSURFACE|SDL::HWSURFACE|SDL::SRCCOLORKEY|SDL::SRCALPHA
      alphaflags = self.flags & (SDL::SRCALPHA|SDL::RLEACCEL)
      
      new_surface=Surface.new(flagbase&self.flags,w,h,self)

      self.set_alpha(0, self.alpha)
      SDL.blitSurface(self,x,y,w,h,new_surface,0,0)
      self.set_alpha(alphaflags, self.alpha)
      
      new_surface.setColorKey(self.flags & (SDL::SRCCOLORKEY|SDL::RLEACCEL),
                              self.colorkey )
      new_surface.setAlpha(self.flags & (SDL::SRCALPHA|SDL::RLEACCEL),
                           self.alpha )
      return new_surface
    end
  end
end

module SGT
  module_function

  def init(w, h, bpp, flag)
    @w = w
    @h = h
    SDL.set_video_mode(w, h, bpp, flag|SDL::OPENGL)
    GL.Viewport(0, 0, w, h)

    GL.Enable(GL::TEXTURE_2D)
    GL::Enable(GL::ALPHA_TEST)
    GL::Color(1.0, 1.0, 1.0, 1.0)

    reset_matrix
  end

  def reset_matrix
    GL.LoadIdentity
    GL.Ortho(0, @w, @h, 0, -1.0, 1.0)
  end

  alias_method :reset_rotzoom, :reset_matrix
  
  def clear(r=0, g=0, b=0)
    GL.ClearColor(r, g, b, 1)
    GL.Clear(GL::COLOR_BUFFER_BIT|GL::DEPTH_BUFFER_BIT)
  end

  def enable_depth_test
    GL.Enable(GL::DEPTH_TEST)
  end

  def disable_depth_test
    GL.Disable(GL::DEPTH_TEST)
  end

  def enable_depth_test?
    GL.IsEnabled(GL::DEPTH_TEST) == GL::TRUE
  end
  
  def enable_alpha_test
    GL.Enable(GL::ALPHA_TEST)
  end

  def disable_alpha_test
    GL.Disable(GL::ALPHA_TEST)
  end

  def enable_alpha_test?
    GL.IsEnabled(GL::ALPHA_TEST) == GL::TRUE
  end
  
  def enable_blend
    GL.Enable(GL::BLEND)
  end
  
  def disable_blend
    GL.Disable(GL::BLEND)
  end

  def enable_blend?
    GL.IsEnabled(GL::BLEND) == GL::TRUE
  end
  
  def set_blend_func(src, dst)
    GL.BlendFunc(src, dst)
  end

  def set_normal_blend
    set_blend_func(Alpha::SRC_ALPHA, Alpha::ONE_MINUS_SRC_ALPHA)
  end

  def set_add_blend
    set_add_blend(Alpha::ONE, Alpha::ONE)
  end

  def depth_mask_on
    GL.DepthMask(GL::TRUE)
  end

  def depth_mask_off
    GL.DepthMask(GL::FALSE)
  end

  def draw_primitive(type, points)
    GL.Disable(GL::TEXTURE_2D)
    GL.Begin(type) do
      points.each{|p| GL.Vertex(p.x, p.y)}
    end
    GL.Enable(GL::TEXTURE_2D)
  end
  
  def draw_line(x1, y1, x2, y2)
    draw_primitive(GL::LINES, [Point.new(x1, y1), Point.new(x2, y2)])
  end

  def draw_rect(x1, y1, x2, y2)
    draw_primitive(GL::LINE_LOOP,
                   [ Point.new(x1, y1),
                     Point.new(x2, y1),
                     Point.new(x2, y2),
                     Point.new(x1, y2)] )

  end

  def draw_strip_lines(points)
    draw_primitive(GL::LINE_STRIP, points)
  end

  def set_color(r, g, b, a=1.0)
    GL.Color(r, g, b, a)
  end
  
  class Error < StandardError
  end

  module Alpha
    ZERO = GL::ZERO
    ONE = GL::ONE
    DST_COLOR = GL::DST_COLOR
    SRC_COLOR = GL::SRC_COLOR
    ONE_MINUS_DST_COLOR = GL::ONE_MINUS_DST_COLOR
    ONE_MINUS_SRC_COLOR = GL::ONE_MINUS_SRC_COLOR
    SRC_ALPHA = GL::SRC_ALPHA
    ONE_MINUS_SRC_ALPHA = GL::ONE_MINUS_SRC_ALPHA
    ONE_MINUS_DST_ALPHA = GL::ONE_MINUS_DST_ALPHA
    SRC_ALPHA_SATURATE = GL::SRC_ALPHA_SATURATE
  end
  
  class Rect
    attr_reader :right, :top, :left, :bottom
    alias x1 left
    alias x2 right
    alias y1 top
    alias y2 bottom
    
    def initialize(left, top, right, bottom)
      @right = right
      @top = top
      @left = left
      @bottom = bottom
    end
    
    def self.new_xywh(x, y, w, h)
      new(x, y, x+w, y+h)
    end

    def relative_of(other)
      Rect.new_xywh(other.left+@left, other.top+@top, self.w, self.h)
    end

    def w
      @right - @left
    end

    def h
      @bottom - @top
    end
    
    def to_texcoord(pseudo_w, pseudo_h)
      Rect.new(@left/Float(pseudo_w), @top/Float(pseudo_h),
               @right/Float(pseudo_w), @bottom/Float(pseudo_h))
    end
  end

  class Point
    attr_reader :x, :y
    def initialize(x, y)
      @x = x
      @y = y
    end
  end

  class EmptyDrawable
    def draw(x, y, z=0)
    end

    def draw_quad(p1, p2, p3, p4, z=0)
    end

    def draw_transform(angle, xscale, yscale, px, py, qx, qy, z=0)
    end
  end
  
  class Drawable
    def draw(x, y, z=0)
      GL::BindTexture(GL::TEXTURE_2D, tex_id)
      GL::Begin(GL::TRIANGLE_FAN) do
        GL::TexCoord(@texcoord_rect.left, @texcoord_rect.top)
        GL::Vertex(x+0.1, y+0.1, z)
        GL::TexCoord(@texcoord_rect.right, @texcoord_rect.top)
        GL::Vertex(x+@rect.w-0.1, y+0.1, z)
        GL::TexCoord(@texcoord_rect.right, @texcoord_rect.bottom)
        GL::Vertex(x+@rect.w-0.1, y+@rect.h-0.1, z)
        GL::TexCoord(@texcoord_rect.left, @texcoord_rect.bottom)
        GL::Vertex(x+0.1, y+@rect.h-0.1, z)
      end
    end

    def draw_quad(p1, p2, p3, p4, z=0)
      GL::BindTexture(GL::TEXTURE_2D, tex_id)
      GL::Begin(GL::TRIANGLE_FAN) do
        GL::TexCoord(@texcoord_rect.left, @texcoord_rect.top)
        GL::Vertex(p1.x, p1.y, z)
        GL::TexCoord(@texcoord_rect.right, @texcoord_rect.top)
        GL::Vertex(p2.x, p2.y, z)
        GL::TexCoord(@texcoord_rect.right, @texcoord_rect.bottom)
        GL::Vertex(p3.x, p3.y, z)
        GL::TexCoord(@texcoord_rect.left, @texcoord_rect.bottom)
        GL::Vertex(p4.x, p4.y, z)
      end
    end

    def draw_transform(angle, xscale, yscale, px, py, qx, qy, z=0)
      GL.PushMatrix
      GL.Translate(qx+0.5, qy+0.5, 0)
      GL.Rotate(angle, 0, 0, 1)
      GL.Scale(xscale, yscale, 1)
      GL.Translate(-(px+0.5), -(py+0.5), 0)
      draw(0, 0, z)
      GL.PopMatrix
    end
  end
  
  class Texture < Drawable
    extend Forwardable
    
    POWER2 = (0..10).map{|i| 2**i}
    LINEAR_FILTER = GL::LINEAR
    NEAREST_FILTER = GL::NEAREST
    
    attr_reader :pseudo_w, :pseudo_h, :tex_id, :rect, :texcoord_rect
    def_delegators :@rect, :w, :h
    
    def self.finalizer(tex_id)
      proc{ GL.DeleteTextures([tex_id]) }
    end
    
    def self.load(fname, alpha=true, colorkey=nil, rect=nil, alpha_test=true)
      surface = SDL::Surface.load(fname)
      
      if rect == nil
        new(surface, alpha, colorkey, alpha_test)
      else
        new(surface.copy_rect(rect.left, rect.top, rect.w, rect.h),
            alpha, colorkey, alpha_test)
      end
    end
    
    def initialize(surface, alpha=true, colorkey=nil, alpha_test=true)
      if surface.w > 1024 || surface.h > 1024
        raise SGT::Error, "Texture size must be smaller than 1024x1024"
      end
      @pseudo_w = round_up_to_pow2(surface.w)
      @pseudo_h = round_up_to_pow2(surface.h)
      @rect = Rect.new(0, 0, surface.w, surface.h)
      @texcoord_rect = @rect.to_texcoord(@pseudo_w, @pseudo_h)
      @texture_alpha = alpha || colorkey
      
      tmp = SDL::Surface.new(SDL::SWSURFACE, @pseudo_w, @pseudo_h, 32,
                             0xff, 0xff << 8, 0xff << 16, 0xff << 24)

      if alpha
        surface.set_alpha(0, 0)
      end
      if !alpha && colorkey
        if colorkey == :default
          surface.set_color_key(SDL::SRCCOLORKEY, surface[0,0])
        else
          surface.set_color_key(SDL::SRCCOLORKEY, colorkey)
        end
        tmp.fill_rect(0, 0, tmp.w, tmp.h, [0, 0, 0, 0])
      end
      tmp.put(surface, 0, 0)

      @tex_id, = GL::GenTextures(1)
      GL::BindTexture(GL::TEXTURE_2D, @tex_id)

      GL::TexParameteri(GL::TEXTURE_2D, GL::TEXTURE_MAG_FILTER, GL::LINEAR)
      GL::TexParameteri(GL::TEXTURE_2D, GL::TEXTURE_MIN_FILTER, GL::NEAREST)

      GL.AlphaFunc(GL::GREATER, 0.003) if @texture_alpha && alpha_test
      
      GL::TexImage2D(GL::TEXTURE_2D, 0, GL::RGBA8, tmp.w, tmp.h, 0,
                     @texture_alpha ? GL::RGBA : GL::RGB,
                     GL::UNSIGNED_BYTE, tmp.pixels)
      ObjectSpace.define_finalizer(self, Texture.finalizer(@tex_id))
    end

    def sub_texture(rect)
      SubTexture.new(self, rect)
    end
    
    def round_up_to_pow2(n)
      POWER2.find{|i| i>=n}
    end
    
    def mag_filter=(filter)
      GL::BindTexture(GL::TEXTURE_2D, @tex_id)
      GL::TexParameteri(GL::TEXTURE_2D, GL::TEXTURE_MAG_FILTER, filter)
    end

    def min_filter=(filter)
      GL::BindTexture(GL::TEXTURE_2D, @tex_id)
      GL::TexParameteri(GL::TEXTURE_2D, GL::TEXTURE_MIN_FILTER, filter)
    end
  end
  
  class SubTexture < Drawable
    extend Forwardable

    def_delegators :@rect, :w, :h
    
    def initialize(parent, rect)
      @parent = parent
      @rect = rect
      @texcoord_rect = rect.to_texcoord(parent.pseudo_w, parent.pseudo_h)
    end
    
    def tex_id
      @parent.tex_id
    end
  end

  class BitmapFont
    def initialize(surface, alpha, colorkey)
      @char_width = surface.w / 256 #/ 
      @characters = (0..255).map do |n|
        Texture.new(surface.copy_rect(n*@char_width, 0, @char_width, surface.h),
                    alpha, colorkey)
      end
      
    end

    def self.load(fname, alpha=true, colorkey=:default)
      surface = SDL::Surface.load(fname)
      if !alpha && colorkey == :default
        return new(surface, false, surface[0,surface.h-1])
      else
        return new(surface, alpha, colorkey)
      end
    end

    def draw(string, x, y, z=0)
      string.enum_for(:each_byte).each_with_index do |ch, n|
        @characters[ch].draw(x+@char_width*n, y, z)
      end
    end
  end

  class SFont
    def self.load(fname, alpha=true, colorkey=:default)
      surface = SDL::Surface.load(fname)
      if !alpha && colorkey == :default
        return new(surface, false, surface[0,0])
      else
        return new(surface, alpha, colorkey)
      end
    end
    
    def initialize(surface, alpha, colorkey)
      xs = search_boundray(surface, 0)
      @characters = Array.new(256, EmptyChar.new(xs[1][0] - xs[0][1]))
      mkchar = proc{|(x1, x2), (x3,x4)|
        Char.new(x1, x2, x3, x4, surface, alpha, colorkey)
      }
      @characters[33,xs.size-1] = xs[0..-2].zip(xs[1..-1]).map(&mkchar)
    end

    def search_boundray(surface, xstart)
      data_color = surface[0,0]
      px = (xstart...surface.w).find{|n| surface[n,0] == data_color}
      qx = (px...surface.w).find{|n| surface[n,0] != data_color}
      if qx.nil? || qx == surface.w - 1
        [[px, surface.w-1]]
      else
        [[px, qx]] + search_boundray(surface, qx)
      end
    end
    
    def draw(string, x, y, z=0)
      string.enum_for(:each_byte).inject(x) do |xx, ch|
        @characters[ch].draw(xx, y,z)
        xx + @characters[ch].adv
      end
    end

    class Char
      attr_reader :adv, :diff, :texture
      
      def initialize(x1, x2, x3, x4, surface, alpha, colorkey)
        @texture = Texture.new(surface.copy_rect((x1+x2)/2,
                                                 1,
                                                 (x4+x3)/2 - (x1+x2)/2,
                                                 surface.h-1),
                               alpha,colorkey) 
        @adv = x3 - x2 
        @diff = (x2 - x1)/2
      end

      def draw(x, y, z)
        @texture.draw(x-@diff, y, z)
      end
    end

    class EmptyChar
      attr_reader :adv
      
      def initialize(adv)
        @adv = adv
      end

      def draw(x, y, z)
      end
    end
  end
end

if __FILE__ == $0
  SDL.init(SDL::INIT_VIDEO)
  SGT.init(640, 480, 24, 0)

  SDL::WM.set_caption("SGT Sample", "SGT Sample")

  texture = SGT::Texture.load("icon.bmp", false, 0)

  p SGT.enable_depth_test?
  p SGT.enable_alpha_test?
  p SGT.enable_alpha_blend?
  
  subtexture = texture.sub_texture(SGT::Rect.new_xywh(32, 32, 16, 16))
  bf = SGT::BitmapFont.load("font.bmp", false)
  sf = SGT::SFont.load("sfont.png", true)
  angle = 0
  
  loop do
    while event = SDL::Event2.poll
      case event
      when SDL::Event2::Quit
        exit
      when SDL::Event2::KeyDown
        exit if event.sym == SDL::Key::ESCAPE
      end
    end

    SGT.clear(0, 0, 0.4)
    SGT.set_color(1, 1, 1)
    
    texture.draw(0, 0)
    texture.draw(34, 0)
    
    subtexture.draw(0, 34)

    texture.draw_quad(SGT::Point.new(80, 5), SGT::Point.new(110, 10),
                      SGT::Point.new(100, 50), SGT::Point.new(70, 50))

    angle += 10
    texture.draw_transform(angle, 1.6, 1.5, 16, 16, 200, 50)

    bf.draw("Now Testing", 5, 80)

    SGT.enable_alpha_blend
    SGT.set_normal_blend
    sf.draw("SFONT Testing", 5, 300)
    SGT.disable_alpha_blend
    
    SGT.set_color(0, 0.5, 0.5)
    SGT.draw_line(20, 100, 120, 200)
    SGT.set_color(1.0, 0, 0.5)
    SGT.draw_rect(20, 100, 120, 200)
    
    SDL.GL_swap_buffers

    sleep 0.05
  end
end
